"""``request`` package contains some classes extending another classes from
``requests`` package to make easier requests to an Airmore server. """
import typing

import ipaddress

import requests.exceptions


class AirmoreRequest(requests.PreparedRequest):
    """A PreparedRequest for Airmore server."""

    def __init__(self, session: "AirmoreSession"):
        super().__init__()
        self.__session = session  # type: AirmoreSession
        self.prepare_method(None)

    def prepare_method(self, method):
        """Will generate "POST" no matter what."""

        self.method = "POST"

    def prepare_url(self, url, params):
        """URL will have ``self._session.base_url``. Only path should be
        provided. For example::

            prepare_url("/foo", {"bar":"baz"})

        will provide::

            self.url  # http://host:port/foo?bar=baz

        depending on your ``self.__session``.
        """

        super().prepare_url(self.__session.base_url+url, params)


class ApplicationOpenRequest(AirmoreRequest):
    """A request to check if application is open."""

    def __init__(self, session: "AirmoreSession"):
        super().__init__(session)
        self.prepare_url("/", {"Key": "PhoneCheckAuthorization"})


class AuthorizationRequest(AirmoreRequest):
    """A request to request application or check if the session is already
    authorized."""

    def __init__(self, session: "AirmoreSession"):
        super().__init__(session)
        self.prepare_url("/", {"Key": "PhoneRequestAuthorization"})


class AirmoreSession(requests.Session):
    """``AirmoreSession`` extends ``requests.Session`` in order to manage an
    Airmore session. """

    def __init__(self, ip_address: ipaddress.IPv4Address, port: int = 2333):
        """
        :param ip_address: IP address to connect to.
        :param port: Port to connect to.
        """
        super().__init__()

        self.ip_address = ip_address  # type: ipaddress.IPv4Address
        self.port = port  # type: int

    @property
    def is_server_running(self) -> bool:
        """Whether the Airmore server runs or not.

        The server is connected via a socket connection.

        :return: True if the server runs.
        """

        import socket
        from contextlib import closing

        is_running = False
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM))\
                as sock:
            if sock.connect_ex((str(self.ip_address), self.port)) == 0:
                is_running = True

        return is_running

    @property
    def is_application_open(self) -> bool:
        """Whether the application is open and front.

        This means the user now can see Airmore application on screen.

        :return: True if the application is open and front.
        """
        # now i know what you are thinking
        # i don't know what these devs were thinking
        # but, apparently, the url below checks if the application
        # is open and user can see it

        request = ApplicationOpenRequest(self)
        response = self.send(request)
        status = response.status_code
        body = response.text

        if status != 200:
            return False

        is_app_front = False

        if body == '"0"':
            is_app_front = True

        return is_app_front

    def request_authorization(self) -> bool:
        """Requests authorization from the device.

        This method will block the thread until the authorization accepted
        on the device. You might want to utilize async if you do not want to
        hang your application.

        The authorization on device will be rejected automatically after 30
        seconds.

        Running this method before each request is a good practice since the
        target device will not show authorization dialog if the session is
        already authorized.

        :return: True if the authorization was accepted.
        """

        request = AuthorizationRequest(self)
        response = self.send(request, False)

        is_accepted = False

        if response.text == "true":
            is_accepted = True

        return is_accepted

    def send(self, request: AirmoreRequest,
             force_authorize: bool = True,
             force_connectivity_check: bool = True,
             **kwargs) -> requests.Response:
        """Sending request with an ``AirmoreRequest``."""

        if force_connectivity_check:
            is_connectivity_present = self.is_server_running

            if not is_connectivity_present:
                raise ServerUnreachableException()

        if force_authorize:
            is_authorized = self.request_authorization()

            if not is_authorized:
                raise AuthorizationException()

        return super().send(request, **kwargs)

    @property
    def base_url(self) -> str:
        """Constructs a base url (prefix) for any AirmoreRequest.

        :return: A base url for internal requests.
        """

        return "http://{}:{}".format(self.ip_address, self.port)


class ServerUnreachableException(Exception):
    """
    This exception is thrown when:
     - No Airmore server was found on target session.
     - Airmore server is idle on target session.
    """

    def __init__(self):
        message = "Server is found idle. The reasons might be:\n"\
                  "- You might not even have installed Airmore to your "\
                  "device.\n"\
                  "- Sometimes your Airmore server goes idle for battery "\
                  "saving. You need to open it up again. "
        super().__init__(message)


class AuthorizationException(Exception):
    """
    This exception is thrown when the authorization is rejected by the device.
    """

    def __init__(self):
        message = "Could not authorize. Please accept authorization on " \
                  "target device. "
        super().__init__(message)


__all__ = [
    "AirmoreSession",
    "AirmoreRequest"
]



"""Service and utilities related to messaging aspect."""
import requests
import uuid

import typing

import datetime
import enum

import pyairmore


class MessageRequestGSMError(Exception):
    """Raised when "/?Key=MessageSend" does not return "2" to imply that
    something bad happened on the server side.
    """

    def __init__(self, *args: object, **kwargs: object) -> None:
        message = "Message could not be sent due to GSM connectivity."
        super().__init__(message, *args, **kwargs)


class MessageType(enum.Enum):
    """Defines if the message was sent or received.

    | **RECEIVED (1)**: Target device received the message.
    | **SENT (2)**: Target device sent the message.
    """

    RECEIVED = 1
    SENT = 2


class Message:  # todo 1 - equality methods
    """A class that consists of information about a particular SMS."""

    def __init__(self):
        self.id = None  # type: str
        self.name = None  # type: str
        self.phone = None  # type: str
        self.datetime = None  # type: datetime.datetime
        self.content = None  # type: str
        self.type = MessageType.RECEIVED  # type: MessageType
        self.was_read = True  # type: bool
        self.count = 1  # type: int

    def __eq__(self, other: "Message") -> bool:
        if self is other:
            return True

        return all((
            self.id == other.id,
            self.name == other.name,
            self.phone == other.phone,
            self.datetime == other.datetime,
            self.content == other.content,
            self.type == other.type,
        ))

    def __ne__(self, other: "Message") -> bool:
        return not self.__eq__(other)

    def __gt__(self, other: "Message") -> bool:
        return self.datetime > other.datetime

    def __ge__(self, other: "Message") -> bool:
        return self.datetime >= other.datetime

    def __lt__(self, other: "Message") -> bool:
        return self.datetime < other.datetime

    def __le__(self, other: "Message") -> bool:
        return self.datetime <= other.datetime


class MessageHistoryRequest(pyairmore.request.AirmoreRequest):
    """A request to get latest messages.

    | **Endpoint:** /?Key=MessageGetLatest
    """

    def __init__(self, session: pyairmore.request.AirmoreSession):
        super().__init__(session)

        self.prepare_url("/", {"Key": "MessageGetLatest"})


class SendMessageRequest(pyairmore.request.AirmoreRequest):
    """A request to send message.

    | **Endpoint:** /?Key=MessageSend
    """

    def __init__(self,
                 session: pyairmore.request.AirmoreSession,
                 phone: str,
                 content: str):
        """
        :param session: Session object, which will be passed to super init.
        :param phone: Phone to send message to.
        :param content: Message content.
        """

        super().__init__(session)

        self.prepare_url("/", {"Key": "MessageSend"})
        self.prepare_headers({})

        data = {
            "Phone": str(phone),
            "Content": str(content),
            "UniqueID": uuid.uuid1().hex
        }
        self.prepare_body("", None, [data])


class ChatHistoryRequest(pyairmore.request.AirmoreRequest):
    """A request to see a particular chat's history.

    | **Endpoint:** /?Key=MessageGetList
    """

    def __init__(self,
                 session: pyairmore.request.AirmoreSession,
                 id: str,
                 start: int = 0,
                 limit: int = 10):
        """
        :param session: Session object, which will be passed to super init.
        :param id: ID of message.
        :param start: Start point.
        :param limit: Limit of messages.
        """

        super().__init__(session)

        self.prepare_url("/", {"Key": "MessageGetList"})
        self.prepare_headers({})

        data = {
            "ID": str(id),
            "Start": int(start),
            "Limit": int(limit)
        }
        self.prepare_body("", None, data)


class MessagingService(pyairmore.services.Service):
    """A service to manage and send messages."""

    def __init__(self, session: pyairmore.request.AirmoreSession):
        super().__init__(session)

    @staticmethod
    def __convert_list_json_to_messages(
            response: requests.Response
    ) -> typing.List[Message]:
        """Will convert a message response from Airmore server to a list of
        ``Message`` objects.
        """

        messages = []
        data = response.json()  # type: typing.List[dict]

        for d in data:
            message = Message()

            message.id = d.get("ID", None)
            message.name = d.get("ShowName", None)
            message.phone = d.get("Phone", None)

            # parsing datetime
            date_str = d.get("Date", None)  # type: str
            if date_str is not None:
                date_str = str(date_str)
                date_format = "%Y/%m/%d %H:%M:%S"

                message.datetime = datetime.datetime.strptime(
                    date_str,
                    date_format
                )

            message.content = d.get("Content", None)

            # detecting type
            message_type = d.get("MsgType", None)
            if message_type and message_type != 1:
                message.type = MessageType.SENT

            # was read
            read = d.get("Read", 1)
            if read != 1:
                message.was_read = False

            message.count = d.get("Count", 1)
            messages.append(message)

        return messages

    def fetch_message_history(self) -> typing.List[Message]:
        """Gets latest messages from your phone. These messages will be
        historically descending order.

        Will return empty list if could not be found.
        """

        request = MessageHistoryRequest(self.session)
        response = self.session.send(request)

        return self.__convert_list_json_to_messages(response)

    def send_message(self,
                     contact_or_phone: typing.Union[str],
                     # todo 3 - support contact
                     content: str) -> None:
        """Sends a single message.

        .. note::

            Contact services will be provided on a future release. Ignore the
            term "contact" in ``contact_or_phone`` parameter.

        :raises MessageRequestGSMError:
        """

        request = SendMessageRequest(self.session, contact_or_phone, content)
        response = self.session.send(request)

        if response.text != "2":
            raise MessageRequestGSMError()

    def fetch_chat_history(self,
                           message_or_id: typing.Union[Message, str],
                           start: int = 0,
                           limit: int = 10) -> typing.List[Message]:
        """Fetches a chat that a particular message is in. These messages will
        be historically descending order.

        Will return empty list if could not be found.

        :param message_or_id: ``Message`` object or a ``str`` id.
        :param start: Starting point.
        :param limit: Limit of messages to fetch.
        """

        if isinstance(message_or_id, Message):
            message_id = message_or_id.id
        else:
            message_id = str(message_or_id)

        start = int(start)
        limit = int(limit)

        request = ChatHistoryRequest(self.session, message_id, start, limit)
        response = self.session.send(request)

        return self.__convert_list_json_to_messages(response)


__all__ = [
    "MessagingService"
]



"""PyAirmore (the package name of which is ``pyairmore``) is a Python client
library for Android Airmore server, which helps the developers to take
programmatic actions on Android. """

import pyairmore.request

__version__ = "0.2.0a1"
__author__ = "Eray Erdin"


def _clean_base64_png(encoded: str) -> str:
    return encoded[22:]



"""Service and utilities related to device aspect."""

import typing

import pyairmore.services
import pyairmore.request


class DeviceDetails:
    """A class that contains several attributes which are details about the
    target device. """

    def __init__(self):
        self.model = None  # type: str
        self.brand = None  # type: str
        self.device_name = None  # type: str
        # phone_number might be "" from airmore server
        self.phone_number = None  # type: str
        self.imei = None  # type: str
        self.imsi = None  # type: str
        self.mac_address = None  # type: str
        self.serial_number = None  # type: str
        self.device_serial_number = None  # type: str
        self.power = 0.0  # type: float
        self.resolution = (0, 0)  # type: typing.Tuple[int]
        self.is_root = False  # type: bool
        self.sdk_version_id = 0  # type: int
        self.sdk_version_name = None  # type: str
        self.platform = 1  # type: int
        self.app_version_code = 1  # type: int
        self.app_version_name = None  # type:str
        self.is_wifi_on = True  # type: bool
        # all in bytes
        self.external_sd_total_size = 0  # type: int
        self.external_sd_available_size = 0  # type: int
        self.internal_sd_total_size = 0  # type: int
        self.internal_sdv_available_size = 0  # type: int
        self.memory_total_size = 0  # type: int
        self.memory_available_size = 0  # type: int
        self.call_history_count = 0  # type: int
        self.contacts_count = 0  # type: int
        self.messages_count = 0  # type: int
        self.pictures_count = 0  # type: int
        self.pictures_total_size = 0  # type: int
        self.songs_count = 0  # type: int
        self.songs_total_size = 0  # type: int
        self.videos_count = 0  # type: int
        self.videos_total_size = 0  # type: int
        self.apks_total_size = 0  # type: int
        self.system_apks_count = 0  # type: int
        self.user_apks_count = 0  # type: int


class DeviceDetailsRequest(pyairmore.request.AirmoreRequest):
    """A request to get device detail.

     | **Endpoint:** /?Key=PhoneGetDeviceInfo&IsDetail=true
    """

    def __init__(self, session: pyairmore.request.AirmoreSession):
        super().__init__(session)

        self.prepare_url("/", {
            "Key": "PhoneGetDeviceInfo",
            "IsDetail": "true"
        })


class DeviceScreenshotRequest(pyairmore.request.AirmoreRequest):
    """A request to get device detail.

     | **Endpoint:** /?Key=PhoneRefreshScreen
    """

    def __init__(self, session: pyairmore.request.AirmoreSession):
        super().__init__(session)

        self.prepare_url("/", {"Key": "PhoneRefreshScreen"})


class DeviceService(pyairmore.services.Service):
    """A service to get details about the target device."""

    def __init__(self, session: pyairmore.request.AirmoreSession):
        super().__init__(session)

    def fetch_device_details(self) -> DeviceDetails:
        """Fetches detail about the target device.

        :return: Detail about the target device.
        """
        request = DeviceDetailsRequest(self.session)
        response = self.session.send(request)
        data = response.json()  # type: dict
        detail = DeviceDetails()

        detail.model = data.get("Model", None)
        detail.brand = data.get("Brand", None)
        detail.device_name = data.get("DeviceName", None)
        # if PhoneNumber is "", then stay None
        detail.phone_number = None if data.get("PhoneNumber", None) == "" \
            else data.get("PhoneNumber", None)
        detail.imei = data.get("IMEI", None)
        detail.imsi = data.get("IMSI", None)
        detail.mac_address = data.get("MAC", None)
        detail.serial_number = data.get("SerialNum", None)
        detail.device_serial_number = data.get("DeviceSN", None)
        detail.power = data.get("Power", 0.0) / 100

        # resolution parsing
        if "*" in data.get("Resolution", ""):
            detail.resolution = tuple(map(int, data.get("Resolution")
                                          .split("*", 1)))

        detail.is_root = False if data.get("Root", None) == 0 else True
        detail.sdk_version_id = data.get("SDKVersionID", 1)
        detail.sdk_version_name = data.get("SDKVersionName", None)
        detail.platform = data.get("Platform", 1)
        detail.app_version_code = data.get("AppVersionCode", 1)
        detail.app_version_name = data.get("AppVersionName", None)
        detail.is_wifi_on = False if data.get("Wifi", None) == 0 else True
        detail.external_sd_total_size = data.get("ExtSDSize", 0)
        detail.external_sd_available_size = data.get("ExtSDAvaSize", 0)
        detail.internal_sd_total_size = data.get("SDSize", 0)
        detail.internal_sdv_available_size = data.get("SDVAvaSize", 0)
        detail.memory_available_size = data.get("MemoryAvaSize", 0)
        detail.memory_total_size = data.get("MemorySize", 0)
        detail.call_history_count = data.get("CallHistoryCount", 0)
        detail.contacts_count = data.get("ContactCount", 0)
        detail.messages_count = data.get("MsgCount", 0)
        detail.pictures_count = data.get("PicCount", 0)
        detail.pictures_total_size = data.get("PicSize", 0)
        detail.songs_count = data.get("MusicCount", 0)
        detail.songs_total_size = data.get("MusicSize", 0)
        detail.videos_count = data.get("VideoCount", 0)
        detail.videos_total_size = data.get("VideoSize", 0)
        detail.apks_total_size = data.get("APKSize", 0)
        detail.system_apks_count = data.get("SystemApkCount", 0)
        detail.user_apks_count = data.get("UserApkCount", 0)

        return detail

    def take_screenshot(self) -> "PIL.Image.Image":
        """Takes screenshot of the target device.

        This method relies on ``pillow`` package. You need to install it.

        :return: Screenshot.
        """
        import PIL.Image

        request = DeviceScreenshotRequest(self.session)
        response = self.session.send(request)
        content = response.text

        from io import BytesIO
        import base64
        from pyairmore import _clean_base64_png

        content = _clean_base64_png(content)
        image = PIL.Image.open(BytesIO(base64.b64decode(content)))

        return image


__all__ = ["DeviceService"]



"""This package contains services to help developers access Airmore server.

Services represent different aspects of Airmore server.
"""

import pyairmore.request


class Service:
    """Service represents an aspect of Airmore, like messages, contacts and
    gallery, each is an aspect and has their services.

    This class is used to extend these aspects. You are likely to use this
    class if you plan to develop/extend the library. If you utilize the
    library, then you need to use other implementations, which are in their
    own modules.

    **Instance Attributes**
     - **session: pyairmore.request.AirmoreSession** -- Parent session.
    """

    def __init__(self, session: pyairmore.request.AirmoreSession):
        """
        :param session: Session to wrap.
        """

        self.session = session
